package com.jakubu9333.grammarManager.api;

import com.fasterxml.jackson.annotation.JsonView;
import com.jakubu9333.grammarManager.FileType;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.View;
import com.jakubu9333.grammarManager.service.GrammarService;
import com.jakubu9333.grammarManager.service.RuleService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.CrossOrigin;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.server.ResponseStatusException;

import java.util.List;

/**
 * @author Jakub Uhlarik
 */
@RestController
@CrossOrigin(origins = "http://localhost:3000")
@RequestMapping(path = "api/grammars")
public class GrammarController {
    public final RuleService ruleService;

    private final GrammarService grammarService;


    @Autowired
    public GrammarController(RuleService ruleService, GrammarService grammarService) {
        this.ruleService = ruleService;
        this.grammarService = grammarService;
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @GetMapping("")
    public ResponseEntity<List<Grammar>> getAllGrammars() {
        return new ResponseEntity<>(grammarService.getGrammars(), HttpStatus.OK);
    }


    @GetMapping("/root/{root}")
    public ResponseEntity<String> getGrammarXmlByRoot(@PathVariable String root) {
        return ResponseEntity.ok().contentType(MediaType.APPLICATION_XML).body(grammarService.getGrammarByRootSymbol(root));
    }

    @GetMapping("name/{name}")
    public ResponseEntity<String> getGrammarByJsgfName(@PathVariable String name){
        return ResponseEntity.ok().contentType(MediaType.TEXT_PLAIN).body(grammarService.getGrammarByJsgfName(name));
    }


    @GetMapping("/{uri}")
    public ResponseEntity<String> getGrammar(@PathVariable String uri, @RequestParam(defaultValue = "xml") String format) {
        FileType typeOfFile;
        switch (format.toLowerCase()) {
            case "xml" -> typeOfFile = FileType.XML_SRGS;
            case "jsgf" -> typeOfFile = FileType.JSGF;
            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "bad chosen format of file");
        }
        String result = grammarService.getGrammarByUri(uri,typeOfFile);
        return ResponseEntity.ok().contentType(typeOfFile == FileType.XML_SRGS ? MediaType.APPLICATION_XML : MediaType.TEXT_PLAIN).body(result);
    }

    @GetMapping("/uri")
    ResponseEntity<String> getAllUris(){
        return ResponseEntity.ok().contentType(MediaType.APPLICATION_JSON).body(grammarService.getAllUris());
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("")
    public ResponseEntity<Grammar> addGrammar(@RequestBody Grammar grammar) {
        return new ResponseEntity<>(grammarService.processAddGrammar(grammar), HttpStatus.CREATED);
    }

    @DeleteMapping("/all")
    public ResponseEntity<?> deleteAllGrammar() {
        grammarService.deleteAll();
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @DeleteMapping("/{id}")
    public ResponseEntity<?> deleteGrammar(@PathVariable Long id) {
        grammarService.deleteById(id);
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("{id}")
    public ResponseEntity<Grammar> duplicateGrammar(@PathVariable Long id) {
        return new ResponseEntity<>(grammarService.duplicateGrammar(id), HttpStatus.CREATED);
    }
}
