package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.FileMetaDataRepo;
import com.jakubu9333.grammarManager.dao.XmlDataDao;
import com.jakubu9333.grammarManager.model.FileHolder;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RuleModel;
import edu.cmu.sphinx.jsgf.JSGFGrammarException;
import edu.cmu.sphinx.jsgf.JSGFGrammarParseException;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammar;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarFactory;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarManager;
import edu.cmu.sphinx.jsgf.parser.JSGFParser;
import edu.cmu.sphinx.jsgf.parser.TokenMgrError;
import edu.cmu.sphinx.jsgf.rule.JSGFRule;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleAlternatives;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleCount;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleName;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleSequence;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleTag;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleToken;
import org.jvoicexml.processor.grammar.Rule;
import org.jvoicexml.processor.grammar.RuleAlternatives;
import org.jvoicexml.processor.grammar.RuleComponent;
import org.jvoicexml.processor.grammar.RuleCount;
import org.jvoicexml.processor.grammar.RuleReference;
import org.jvoicexml.processor.grammar.RuleSequence;
import org.jvoicexml.processor.grammar.RuleSpecial;
import org.jvoicexml.processor.grammar.RuleTag;
import org.jvoicexml.processor.grammar.RuleToken;
import org.jvoicexml.processor.srgs.GrammarException;
import org.jvoicexml.processor.srgs.xml.SrgsRuleGrammarParser;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.transaction.Transactional;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Jakub Uhlarik
 */
@Service
public class FileService {

    private final FileMetaDataRepo fileMetaDataRepo;

    private final XmlDataDao xmlDataDao;

    @Autowired
    public FileService(FileMetaDataRepo fileMetaDataRepo, XmlDataDao xmlDataDao) {
        this.xmlDataDao = xmlDataDao;
        this.fileMetaDataRepo = fileMetaDataRepo;
    }


    public static class XmlGrammarHandler extends DefaultHandler {
        private Grammar newGrammar;
        private StringBuilder rightSide;

        private RuleModel rule = null;

        private boolean grammar = false;

        @Override
        public void startDocument() {
            newGrammar = new Grammar(null, new HashSet<>(), null);
        }

        @Override
        public void characters(char[] ch, int start, int length) {
            if (rightSide != null) {
                rightSide.append(ch, start, length);
            }
        }

        @Override
        public void startElement(String uri, String lName, String qName, Attributes attr) {
            switch (qName) {
                case "grammar" -> {
                    grammar = true;
                    String grammarUri = attr.getValue("xml:base");
                    if (grammarUri != null) {
                        newGrammar.setUri(grammarUri);
                    }
                    String rootLeft = attr.getValue("root");
                    if (rootLeft != null) {
                        newGrammar.setRootSymbol(rootLeft);
                    }
                    return;
                }
                case "rule" -> {
                    if (!grammar) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "rule outside grammar");
                    }
                    rule = new RuleModel(newGrammar, attr.getValue("id"), null, false);
                    rightSide = new StringBuilder();
                    return;
                }
                case "item" -> {
                    if (rule == null) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "item outside rule");
                    }
                }
                case "one-of" -> {
                    if (rule == null) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "one-of outside rule");
                    }
                }
            }


            if (rule != null) {
                rightSide.append("<").append(qName);

                for (int i = 0; i < attr.getLength(); i++) {
                    rightSide.append(" ").append(attr.getQName(i)).append("=\"").append(attr.getValue(i)).append('"');
                }

                rightSide.append(">");


            }


        }

        @Override
        public void endElement(String uri, String lName, String qName) {

            if ("rule".equals(qName)) {
                rule.setRightSide(rightSide.toString());

                newGrammar.getRules().add(rule);
                rule = null;
                return;
            }
            if (rule != null) {
                rightSide.append("</").append(qName).append(">");
            }
        }
    }

    public FileHolder getFileFromGrammar(Long id) {
        return fileMetaDataRepo.findByGrammarId(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.BAD_REQUEST, "no grammar found"));
    }


    //function validates xml document and transforms jsgf document to xml
    private String parseFile(String file, boolean xml) {

        if (xml) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();

                XmlGrammarHandler handler = new XmlGrammarHandler();
                InputSource is = new InputSource(new StringReader(file));
                saxParser.parse(is, handler);
                return file;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "xml parse error\n" + e.getMessage(), e);
            }

        }
        try {
            return jsgfToXml(file);
        } catch (JSGFGrammarParseException | TokenMgrError | IOException | URISyntaxException |
                 JSGFGrammarException e) {
            if (e instanceof JSGFGrammarParseException jsgfGrammarParseException) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + jsgfGrammarParseException.details);
            }
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + e.getMessage(), e);
        }


    }


    private boolean getFileExtension(String file) {
        Pattern xml = Pattern.compile("^\\W*<\\?xml.*\\?>");
        Matcher mXml = xml.matcher(file);
        if (mXml.find()) {
            return true;
        }
        Pattern jsgf = Pattern.compile("^\\W*#JSGF.*;");
        Matcher mJsgf = jsgf.matcher(file);
        if (mJsgf.find()) {
            return false;
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "format cant be found from file");
    }


    @Transactional
    public String addGrammarFromFile(String file, Boolean xml) {
        if (xml == null) {
            xml = getFileExtension(file);

        }
        String grammar = parseFile(file, xml);
        return xmlDataDao.add(grammar);
    }

    private List<JSGFRule> makeSequenceSmaller(JSGFRuleSequence jsgfRuleSequence) {
        List<JSGFRule> reducedList = new ArrayList<>();
        StringBuilder sb = new StringBuilder();
        for (JSGFRule ruleInSequence : jsgfRuleSequence.getRules()) {
            if (ruleInSequence instanceof JSGFRuleToken jsgfRuleToken) {
                if (!sb.isEmpty()) {
                    sb.append(" ");
                }
                sb.append(jsgfRuleToken.getText());
            } else {
                if (!sb.isEmpty()) {
                    reducedList.add(new JSGFRuleToken(sb.toString()));
                    sb = new StringBuilder();
                }
                reducedList.add(ruleInSequence);
            }
        }
        if (!sb.isEmpty()) {
            reducedList.add(new JSGFRuleToken(sb.toString()));
        }
        return reducedList;
    }

    private RuleComponent jsgfToXml(JSGFRule jsgfRule) {
        if (jsgfRule instanceof JSGFRuleAlternatives jsgfRuleAlternatives) {
            if (jsgfRuleAlternatives.getRules().size() == 1) {
                return jsgfToXml(jsgfRuleAlternatives.getRules().get(0));
            }
            RuleAlternatives ruleAlternatives = new RuleAlternatives();
            List<Float> weights = jsgfRuleAlternatives.getWeights();
            AtomicInteger i = new AtomicInteger();
            jsgfRuleAlternatives.getRules().forEach(p -> {
                if (weights == null) {
                    ruleAlternatives.addAlternative(jsgfToXml(p));
                } else {
                    ruleAlternatives.addAlternative(jsgfToXml(p), weights.get(i.get()));
                    i.getAndIncrement();
                }
            });
            return ruleAlternatives;
        }
        if (jsgfRule instanceof JSGFRuleSequence jsgfRuleSequence) {
            if (jsgfRuleSequence.getRules().size() == 1) {
                return jsgfToXml(jsgfRuleSequence.getRules().get(0));
            }
            //todo
            List<JSGFRule> reducedList = makeSequenceSmaller(jsgfRuleSequence);
            if (reducedList.size() == 1) {
                return jsgfToXml(reducedList.get(0));
            }
            RuleSequence ruleSequence = new RuleSequence();

            reducedList.forEach(p -> ruleSequence.addElement(jsgfToXml(p)));
            return ruleSequence;
        }
        if (jsgfRule instanceof JSGFRuleCount jsgfRuleCount) {
            int repeat = jsgfRuleCount.getCount();
            RuleComponent ruleComponent = jsgfToXml(jsgfRuleCount.getRule());
            RuleCount ruleCount;
            switch (repeat) {
                case JSGFRuleCount.ZERO_OR_MORE -> ruleCount = new RuleCount(ruleComponent, 0);
                case JSGFRuleCount.OPTIONAL -> ruleCount = new RuleCount(ruleComponent, 0, 1);
                case JSGFRuleCount.ONCE_OR_MORE -> ruleCount = new RuleCount(ruleComponent, 1);
                default -> ruleCount = new RuleCount(ruleComponent, 1, 1);
            }
            return ruleCount;
        }
        if (jsgfRule instanceof JSGFRuleName jsgfRuleName) {
            if (jsgfRuleName.getRuleName().equals("NULL")) {
                return RuleSpecial.NULL;
            }
            if (jsgfRuleName.getRuleName().equals("VOID")) {
                return RuleSpecial.VOID;
            }
            return new  RuleReference(jsgfRuleName.getRuleName());
        }
        if (jsgfRule instanceof JSGFRuleToken jsgfRuleToken) {
            return new RuleToken(jsgfRuleToken.getText());
        }
        if (jsgfRule instanceof JSGFRuleTag jsgfRuleTag) {
            return new RuleTag(jsgfRuleTag.getTag());
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "unknown node");


    }

    public String XmltoJSGF(String xml, boolean grammar) {
        StringBuilder xmlString = new StringBuilder();
        SrgsRuleGrammarParser parser = new SrgsRuleGrammarParser();
        writeHeader(xmlString, null, true, grammar);
        xmlString.append(xml);
        if (!grammar) {
            xmlString.append("</grammar>");
        }

        List<Rule> rules;
        try {
            rules = parser.load(new StringReader(xmlString.toString()));

        } catch (URISyntaxException | GrammarException e) {
            throw new ResponseStatusException(HttpStatus.INTERNAL_SERVER_ERROR, "error in converting xml to jsgf");
        }
        StringBuilder jsgfString = new StringBuilder();
        if (grammar) {
            writeHeader(jsgfString, (String) parser.getAttributes().get("name"), false,false);
            Object imports = parser.getAttributes().get("imports");
            if (imports != null) {
                jsgfString.append(imports);
            }
        }

        rules.forEach(p -> jsgfString.append(p.toStringJSGF()).append("\n"));
        return jsgfString.toString();
    }

    private String jsgfToXml(String file) throws JSGFGrammarParseException, IOException, URISyntaxException, JSGFGrammarException {
        JSGFRuleGrammar grammarRaw = JSGFParser.newGrammarFromJSGF(new StringReader(file), new JSGFRuleGrammarFactory(new JSGFRuleGrammarManager()));
        StringBuilder sb = new StringBuilder();
        writeHeader(sb, null, true,false);
        for (JSGFRuleName importName : grammarRaw.getImports()) {
            sb.append("<meta name='jsgfImport' content='").append(importName.getRuleName()).append("'").append("/>\n");
        }
        sb.append("<meta name='jsgfName' content='").append(grammarRaw.getName()).append("'/>\n");
        for (String ruleName : grammarRaw.getRuleNames()) {
            JSGFRule rule = grammarRaw.getRule(ruleName);
            boolean publicRule =grammarRaw.isRulePublic(ruleName);
            writeXmlRule(sb, publicRule,ruleName, jsgfToXml(rule));
        }
        sb.append("</grammar>");

        return sb.toString();
    }

    private void writeXmlRule(StringBuilder sb,boolean publicRule, String name, RuleComponent ruleComponent) {
        sb.append("<rule ");
        if (publicRule){
            sb.append("scope=\"public\" ");
        }
        sb.append("id=\"");
        sb.append(name).append("\">\n");
        sb.append(ruleComponent.toStringXML());
        sb.append("</rule>\n");
    }


    private void writeHeader(StringBuilder sb, String name, boolean xml, boolean justPrefix) {
        //jsgf
        if (!xml) {
            sb.append("#JSGF  V1.0;\n");
            sb.append("grammar ").append(name == null ? "" : name).append(";\n\n");
            return;
        }

        //xml
        sb.append("<?xml version=\"1.0\"?>\n");
        if (!justPrefix) {
            sb.append("<grammar version=\"1.0\"\n");
            sb.append("xml:lang='cz'");
            sb.append(" xmlns=\"http://www.w3.org/2001/06/grammar\">\n");
        }

    }

}

