package com.jakubu9333.grammarManager.service;



import com.jakubu9333.grammarManager.FileType;
import com.jakubu9333.grammarManager.dao.GrammarRepository;

import com.jakubu9333.grammarManager.dao.XmlDataDao;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RuleModel;
import org.springframework.beans.factory.annotation.Autowired;

import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;


import javax.transaction.Transactional;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;


import java.io.IOException;
import java.io.StringReader;
import java.util.HashSet;
import java.util.List;

import java.util.Objects;
import java.util.Set;

import static java.lang.Math.abs;


/**
 * @author Jakub Uhlarik
 */
@Service
public class GrammarService {

    private final GrammarRepository grammarRepository;
    private final XmlDataDao xmlDataDao;
    private final FileService fileService;






    @Autowired
    public GrammarService(GrammarRepository grammarRepository,
                          XmlDataDao xmlDataDao, FileService fileService) {
        this.grammarRepository = grammarRepository;
        this.xmlDataDao = xmlDataDao;
        this.fileService = fileService;
    }

    public List<Grammar> getGrammars() {
        return grammarRepository.findAll();
    }


    @Transactional
    public void deleteAll() {
        grammarRepository.deleteAll();
    }


    @Transactional
    public void deleteById(Long id) {
        grammarRepository.deleteById(id);
    }

    public Grammar preProcessGrammar(Grammar grammar) {
        grammar.setId(null);
        if (grammar.getUri() == null) {
            grammar.setUri("null");
        }

        String rootRule = grammar.getRootSymbol();
        Set<RuleModel> rules = grammar.getRules();
        if (rootRule != null) {
            //ads root attribute to rule in rules
            rules.forEach(p -> {
                if (Objects.equals(p.getLeftSymbol(), rootRule)) {
                    p.setRoot(true);
                }
                p.setGrammar(grammar);
            });
        }


        return grammar;
    }

    @Transactional
    public Grammar processAddGrammar(Grammar grammar) {
        return addGrammar(preProcessGrammar(grammar));
    }

    @Transactional
    public Grammar addGrammar(Grammar grammar) {
        grammar = grammarRepository.save(grammar);
        return grammar;
    }

    public Grammar getGrammarById(Long id) {
        return grammarRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "grammar id: " + id + " not found"));
    }

    @Transactional
    public Grammar duplicateGrammar(Long id) {
        Grammar grammar = getGrammarById(id);
        Grammar newGrammar = new Grammar();
        newGrammar.setUri(grammar.getUri());
        newGrammar.setRootSymbol(grammar.getRootSymbol());
        Set<RuleModel> rules = grammar.getRules();
        Set<RuleModel> newRules = new HashSet<>();
        for (RuleModel rule : rules) {
            newRules.add(new RuleModel(rule));
        }
        newGrammar.setRules(newRules);
        newGrammar = addGrammar(newGrammar);
        return newGrammar;
    }

    public String getGrammarByRootSymbol(String rootSymbol) {
        String result = xmlDataDao.getByRoot(rootSymbol);
        if (result.isEmpty()){
            throw  new ResponseStatusException(HttpStatus.NOT_FOUND,"no grammar with "+rootSymbol+" rootsymbol");
        }
        return result;
    }

    public String getAllUris(){
        return xmlDataDao.allUris();
    }


    public String deleteGrammar(String uri){
        return xmlDataDao.delete(uri);
    }


    public String getGrammarByUri(String uri, FileType typeOfFile) {
        String result =  xmlDataDao.getByUri(uri);
        if( result.isEmpty()){
            throw new ResponseStatusException(HttpStatus.NOT_FOUND,"grammar with "+uri+" not found");
        }
        if (typeOfFile==FileType.JSGF){
            return fileService.XmltoJSGF(result,true);
        }
        return result;

    }

    public String getGrammarByJsgfName(String name){
        String jsgfGrammar =  xmlDataDao.getByJsgfName(name);
        if (jsgfGrammar.isEmpty()){
            throw new ResponseStatusException(HttpStatus.NOT_FOUND,"grammar with name "+name+" not found");
        }

        return fileService.XmltoJSGF(jsgfGrammar,true);
    }


}
